# Copyright (c) 2024 Lucas Dooms

# File containing different initial DNA configurations

from __future__ import annotations

from dataclasses import dataclass
from itertools import product
from typing import Dict, List, Tuple

import numpy as np

from smc_lammps.generate.default_parameters import Parameters
from smc_lammps.generate.generator import (
    BAI,
    AtomGroup,
    AtomIdentifier,
    AtomType,
    BAI_Kind,
    BAI_Type,
    MoleculeId,
    Nx3Array,
    PairWise,
)
from smc_lammps.generate.structures import structure_creator
from smc_lammps.generate.structures.dna import dna_creator
from smc_lammps.generate.structures.smc.smc import SMC
from smc_lammps.generate.util import get_closest, pos_from_id


@dataclass
class DnaParameters:
    nDNA: int
    DNA_bond_length: float
    DNA_mass: float
    type: AtomType
    mol_DNA: int
    bond: BAI_Type
    angle: BAI_Type
    ssangle: BAI_Type

    def create_dna(self, dna_positions) -> List[AtomGroup]:
        return [
            AtomGroup(
                positions=r_DNA,
                atom_type=self.type,
                molecule_index=self.mol_DNA,
                polymer_bond_type=self.bond,
                polymer_angle_type=self.angle,
            )
            for r_DNA in dna_positions
        ]


@dataclass
class InteractionParameters:
    ###########
    # DNA-DNA #
    ###########

    sigma_DNA_DNA: float
    epsilon_DNA_DNA: float
    rcut_DNA_DNA: float

    ###########
    # SMC-DNA #
    ###########

    sigma_SMC_DNA: float
    epsilon_SMC_DNA: float
    rcut_SMC_DNA: float

    #############
    # Sites-DNA #
    #############

    # Sigma of LJ attraction (same as those of the repulsive SMC sites)
    sigma_upper_site_DNA: float

    # Cutoff distance of LJ attraction
    rcut_lower_site_DNA: float

    # Epsilon parameter of LJ attraction
    epsilon_upper_site_DNA: float


@dataclass
class Tether:
    class Obstacle:
        def move(self, vector) -> None:
            raise NotImplementedError("don't use Tether.Obstacle directly")

    class Wall(Obstacle):
        def __init__(self, y_pos: float) -> None:
            super().__init__()
            self.y_pos = y_pos

        def move(self, vector) -> None:
            self.y_pos += vector[1]

    class Gold(Obstacle):
        def __init__(
            self, group: AtomGroup, radius: float, cut: float, tether_bond: BAI
        ) -> None:
            super().__init__()
            self.group = group
            self.radius = radius
            self.cut = cut
            self.tether_bond = tether_bond

        def move(self, vector) -> None:
            self.group.positions[0] += vector

    group: AtomGroup
    dna_tether_id: AtomIdentifier
    obstacle: Tether.Obstacle

    @staticmethod
    def get_gold_mass(radius: float) -> float:
        """radius in nanometers, returns attograms"""
        density = 0.0193  # attograms per nanometer^3
        volume = 4.0 / 3.0 * np.pi * radius**3
        return density * volume

    @classmethod
    def get_obstacle(
        cls, real_obstacle: bool, ip: InteractionParameters, tether_group: AtomGroup
    ) -> Tether.Obstacle:
        if real_obstacle:
            obstacle_radius = 100  # nanometers
            obstacle_cut = obstacle_radius * 2 ** (1 / 6)
            pos = tether_group.positions[0] - np.array(
                [0, obstacle_radius - ip.sigma_DNA_DNA, 0], dtype=float
            )
            obstacle_type = AtomType(cls.get_gold_mass(obstacle_radius))
            obstacle_group = AtomGroup(
                positions=np.array([pos]),
                atom_type=obstacle_type,
                molecule_index=tether_group.molecule_index,
            )

            obstacle_bond = BAI_Type(
                BAI_Kind.BOND,
                "fene/expand",
                f"{1.0} {obstacle_radius} {0.0} {0.0} {ip.sigma_DNA_DNA}\n",
            )
            tether_obstacle_bond = BAI(
                obstacle_bond, (tether_group, 0), (obstacle_group, 0)
            )
            return Tether.Gold(
                obstacle_group, obstacle_radius, obstacle_cut, tether_obstacle_bond
            )
        else:
            return Tether.Wall(tether_group.positions[0][1])

    @classmethod
    def create_tether(
        cls,
        dna_tether_id: AtomIdentifier,
        tether_length: int,
        bond_length: float,
        mass: float,
        bond_type: BAI_Type,
        angle_type: BAI_Type,
        obstacle: Tether.Obstacle,
    ) -> Tether:
        tether_positions = (
            structure_creator.get_straight_segment(tether_length, [0, 1, 0])
            * bond_length
        )
        tether_group = AtomGroup(
            positions=tether_positions,
            atom_type=AtomType(mass),
            molecule_index=MoleculeId.get_next(),
            polymer_bond_type=bond_type,
            polymer_angle_type=angle_type,
            charge=0.2,
        )

        return Tether(
            group=tether_group, dna_tether_id=dna_tether_id, obstacle=obstacle
        )

    def move(self, vector) -> None:
        self.group.positions += vector
        self.obstacle.move(vector)

    def get_all_groups(self) -> List[AtomGroup]:
        groups = [self.group]
        if isinstance(self.obstacle, Tether.Gold):
            groups += [self.obstacle.group]
        return groups

    def handle_end_points(self, end_points: List[AtomIdentifier]) -> None:
        # freeze bottom of tether if using infinite wall
        if isinstance(self.obstacle, Tether.Wall):
            end_points += [(self.group, 0)]

    def add_interactions(
        self,
        pair_inter: PairWise,
        ip: InteractionParameters,
        dna_type: AtomType,
        smc: SMC,
        kBT: float,
    ) -> None:
        tether_type = self.group.type
        # tether
        pair_inter.add_interaction(
            tether_type,
            tether_type,
            ip.epsilon_DNA_DNA * kBT,
            ip.sigma_DNA_DNA,
            ip.rcut_DNA_DNA,
        )
        pair_inter.add_interaction(
            tether_type,
            dna_type,
            ip.epsilon_DNA_DNA * kBT,
            ip.sigma_DNA_DNA,
            ip.rcut_DNA_DNA,
        )
        pair_inter.add_interaction(
            tether_type,
            smc.t_arms_heads_kleisin,
            ip.epsilon_SMC_DNA * kBT,
            ip.sigma_SMC_DNA,
            ip.rcut_SMC_DNA,
        )
        pair_inter.add_interaction(
            tether_type,
            smc.t_hinge,
            ip.epsilon_SMC_DNA * kBT,
            ip.sigma_SMC_DNA / 2.0,
            ip.rcut_SMC_DNA / 2.0,
        )
        # Optional: don't allow bridge to go through tether
        # pair_inter.add_interaction(
        #     tether_type, smc.atp_type,
        #     ip.epsilonSMCvsDNA * kBT, ip.sigmaSMCvsDNA, ip.rcutSMCvsDNA
        # )
        # Optional: allow tether to bond to siteD
        # pair_inter.add_interaction(
        #     tether_type, smc.siteD_type,
        #     ip.epsilonSiteDvsDNA * kBT, ip.sigmaSiteDvsDNA, ip.rcutSiteDvsDNA
        # )
        if isinstance(self.obstacle, Tether.Gold):
            pair_inter.add_interaction(
                self.obstacle.group.type,
                dna_type,
                ip.epsilon_DNA_DNA * kBT,
                self.obstacle.radius,
                self.obstacle.cut,
            )
            pair_inter.add_interaction(
                self.obstacle.group.type,
                smc.t_arms_heads_kleisin,
                ip.epsilon_DNA_DNA * kBT,
                self.obstacle.radius,
                self.obstacle.cut,
            )
            pair_inter.add_interaction(
                self.obstacle.group.type,
                smc.t_hinge,
                ip.epsilon_DNA_DNA * kBT,
                self.obstacle.radius,
                self.obstacle.cut,
            )
            pair_inter.add_interaction(
                self.obstacle.group.type,
                tether_type,
                ip.epsilon_DNA_DNA * kBT,
                self.obstacle.radius,
                self.obstacle.cut,
            )

    def get_bonds(self, bond_type: BAI_Type) -> List[BAI]:
        bonds = [BAI(bond_type, (self.group, -1), self.dna_tether_id)]
        if isinstance(self.obstacle, Tether.Gold):
            bonds += [self.obstacle.tether_bond]
        return bonds


# decorator to add tether logic to DnaConfiguration classes
def with_tether(cls):
    def new1(f):
        def get_all_groups(self) -> List[AtomGroup]:
            return f(self) + self.tether.get_all_groups()

        return get_all_groups

    cls.get_all_groups = new1(cls.get_all_groups)

    def new2(f):
        def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
            ppp = f(self)
            self.tether.handle_end_points(ppp.end_points)
            return ppp

        return get_post_process_parameters

    cls.get_post_process_parameters = new2(cls.get_post_process_parameters)

    def new3(f):
        def add_interactions(self, pair_inter: PairWise) -> None:
            f(self, pair_inter)
            self.tether.add_interactions(
                pair_inter,
                self.inter_par,
                self.dna_parameters.type,
                self.smc,
                self.par.kB * self.par.T,
            )

        return add_interactions

    cls.add_interactions = new3(cls.add_interactions)

    def new4(f):
        def get_bonds(self) -> List[BAI]:
            return f(self) + self.tether.get_bonds(self.dna_parameters.bond)

        return get_bonds

    cls.get_bonds = new4(cls.get_bonds)

    return cls


class DnaConfiguration:
    @dataclass
    class PostProcessParameters:
        # LAMMPS DATA

        # indices to freeze permanently
        end_points: List[AtomIdentifier]
        # indices to temporarily freeze, in order to equilibrate the system
        freeze_indices: List[AtomIdentifier]
        # forces to apply:
        # the keys are the forces (3d vectors), and the value is a list of indices to which the force will be applied
        stretching_forces_array: Dict[Tuple[float, float, float], List[AtomIdentifier]]

        # POST PROCESSING

        # indices to use for marked bead tracking
        dna_indices_list: List[Tuple[AtomIdentifier, AtomIdentifier]]

    @classmethod
    def set_parameters(cls, par: Parameters, inter_par: InteractionParameters) -> None:
        cls.par = par
        cls.inter_par = inter_par

    @classmethod
    def set_smc(cls, smc: SMC) -> None:
        cls.smc = smc

    def __init__(
        self, dna_groups: List[AtomGroup], dna_parameters: DnaParameters
    ) -> None:
        self.dna_groups = dna_groups
        self.dna_parameters = dna_parameters
        self.kBT = self.par.kB * self.par.T
        self.beads: List[AtomGroup] = []
        self.bead_sizes: List[float] = []
        self.bead_bonds: List[BAI] = []

    def get_all_groups(self) -> List[AtomGroup]:
        return self.dna_groups + self.beads

    @property
    def dna_full_list(self) -> Nx3Array:
        return np.concatenate([grp.positions for grp in self.dna_groups])

    @property
    def dna_full_list_length(self) -> int:
        return sum(len(grp.positions) for grp in self.dna_groups)

    def get_dna_id_from_list_index(self, index: int) -> AtomIdentifier:
        if index < 0:
            index += self.dna_full_list_length
        assert index >= 0

        for grp in self.dna_groups:
            if index < len(grp.positions):
                return (grp, index)
            index -= len(grp.positions)
        raise IndexError(f"index {index} out of bounds for DNA groups.")

    @property
    def first_dna_id(self) -> AtomIdentifier:
        return self.get_dna_id_from_list_index(0)

    @property
    def last_dna_id(self) -> AtomIdentifier:
        return self.get_dna_id_from_list_index(-1)

    def get_percent_dna_id(self, ratio: float) -> AtomIdentifier:
        return self.get_dna_id_from_list_index(int(ratio * self.dna_full_list_length))

    @classmethod
    def get_dna_config(
        cls, dna_parameters: DnaParameters, r_lower_site, par
    ) -> DnaConfiguration:
        return NotImplemented

    def get_post_process_parameters(self) -> PostProcessParameters:
        return self.PostProcessParameters(
            end_points=[],
            freeze_indices=[],
            stretching_forces_array=dict(),
            dna_indices_list=[],
        )

    def dna_indices_list_get_all_dna(
        self,
    ) -> List[Tuple[AtomIdentifier, AtomIdentifier]]:
        return [((dna_grp, 0), (dna_grp, -1)) for dna_grp in self.dna_groups]

    def dna_indices_list_get_dna_to(
        self, ratio: float
    ) -> List[Tuple[AtomIdentifier, AtomIdentifier]]:
        return [
            ((dna_grp, 0), (dna_grp, int(len(dna_grp.positions) * ratio)))
            for dna_grp in self.dna_groups
        ]

    def add_interactions(self, pair_inter: PairWise) -> None:
        dna_type = self.dna_parameters.type
        ip = self.inter_par
        kBT = self.par.kB * self.par.T
        pair_inter.add_interaction(
            dna_type,
            dna_type,
            ip.epsilon_DNA_DNA * kBT,
            ip.sigma_DNA_DNA,
            ip.rcut_DNA_DNA,
        )
        pair_inter.add_interaction(
            dna_type,
            self.smc.t_arms_heads_kleisin,
            ip.epsilon_SMC_DNA * kBT,
            ip.sigma_SMC_DNA,
            ip.rcut_SMC_DNA,
        )
        pair_inter.add_interaction(
            dna_type,
            self.smc.t_hinge,
            ip.epsilon_SMC_DNA * kBT,
            ip.sigma_SMC_DNA,
            ip.rcut_SMC_DNA,
        )
        pair_inter.add_interaction(
            dna_type,
            self.smc.t_lower_site,
            ip.epsilon_upper_site_DNA * kBT,
            ip.sigma_upper_site_DNA,
            ip.rcut_lower_site_DNA,
        )

        # every bead should repel every SMC group
        for (bead, bead_size), smc_grp in product(
            zip(self.beads, self.bead_sizes), self.smc.get_groups()
        ):
            pair_inter.add_interaction(
                bead.type,
                smc_grp.type,
                ip.epsilon_SMC_DNA * kBT,
                bead_size,
                bead_size * (2 ** (1 / 6)),
            )

    def get_bonds(self) -> List[BAI]:
        return self.bead_bonds

    def update_tether_bond(
        self, old_id: AtomIdentifier, new_groups, bead: None | AtomIdentifier
    ) -> None:
        if not hasattr(self, "tether"):
            return
        assert isinstance(self.tether, Tether)

        if self.tether.dna_tether_id[0] is old_id[0]:
            if self.tether.dna_tether_id[1] < old_id[1]:
                self.tether.dna_tether_id = (
                    new_groups[0],
                    self.tether.dna_tether_id[1],
                )
            elif self.tether.dna_tether_id[1] == old_id[1] and bead is not None:
                self.tether.dna_tether_id = bead
            else:
                self.tether.dna_tether_id = (
                    new_groups[1],
                    self.tether.dna_tether_id[1] - old_id[1],
                )

        old = pos_from_id(old_id)
        new = pos_from_id(self.tether.dna_tether_id)
        self.tether.move(new - old)

    def split_dna(self, split: AtomIdentifier) -> Tuple[AtomGroup, AtomGroup]:
        """split DNA in two pieces, with the split atom id part of the second group."""
        self.dna_groups.remove(split[0])
        pos1 = split[0].positions[: split[1]]
        pos2 = split[0].positions[split[1] :]

        args = (
            split[0].type,
            split[0].molecule_index,
            split[0].polymer_bond_type,
            split[0].polymer_angle_type,
        )
        groups = (
            AtomGroup(pos1, *args),
            AtomGroup(pos2, *args),
        )
        self.dna_groups += groups
        return groups

    def add_bead_to_dna(
        self,
        bead_type: AtomType,
        mol_index: int,
        dna_atom: AtomIdentifier,
        bond: None | BAI_Type,  # if None -> rigid attachment to dna_atom
        angle: None | BAI_Type,  # only used if bond is not None
        bead_size: float,
    ) -> None:
        # place on a DNA bead
        location = pos_from_id(dna_atom)

        # create a bead
        bead = AtomGroup(location.reshape(1, 3), bead_type, mol_index)

        bais = []
        if bond is None:
            # TODO:
            # gen.molecule_override[dna_atom] = mol_index
            pass
        else:
            first_group, second_group = self.split_dna(dna_atom)

            # add interactions/exceptions
            bais += [
                BAI(bond, (first_group, -1), (bead, 0)),
                BAI(bond, (second_group, 0), (bead, 0)),
            ]
            if angle is not None:
                bais += [
                    BAI(angle, (first_group, -2), (first_group, -1), (bead, 0)),
                    BAI(angle, (first_group, -1), (bead, 0), (second_group, 0)),
                    BAI(angle, (bead, 0), (second_group, 0), (second_group, 1)),
                ]

            # move to correct distances
            bead.positions[0, 0] += bead_size
            first_group.positions[:, 0] += (
                2 * bead_size - self.dna_parameters.DNA_bond_length
            )

            self.update_tether_bond(dna_atom, (first_group, second_group), (bead, 0))

        self.beads.append(bead)
        self.bead_sizes.append(bead_size)
        self.bead_bonds += bais

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        return []

    @staticmethod
    def str_to_config(string: str):
        string = string.lower()
        return {
            "line": Line,
            "folded": Folded,
            "right_angle": RightAngle,
            "doubled": Doubled,
            "safety": Safety,
            "obstacle": Obstacle,
            "obstacle_safety": ObstacleSafety,
            "advanced_obstacle_safety": AdvancedObstacleSafety,
        }[string]


class Line(DnaConfiguration):
    """Straight line of DNA"""

    def __init__(self, dna_groups, dna_parameters: DnaParameters):
        super().__init__(dna_groups, dna_parameters)

    @classmethod
    def get_dna_config(cls, dna_parameters: DnaParameters, r_lower_site, par) -> Line:
        default_dna_pos = r_lower_site[1] + np.array([0, par.cutoff6, 0])

        # 1.
        [r_DNA] = dna_creator.get_dna_coordinates_straight(
            dna_parameters.nDNA, dna_parameters.DNA_bond_length
        )

        # 2.
        # make sure SMC contains DNA
        goal = default_dna_pos
        start = np.array(
            [
                r_DNA[int(len(r_DNA) / 1.3)][0] + 10.0 * dna_parameters.DNA_bond_length,
                r_DNA[-1][1],
                0,
            ]
        )
        shift = (goal - start).reshape(1, 3)
        r_DNA += shift

        return cls(dna_parameters.create_dna([r_DNA]), dna_parameters)

    def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
        ppp = super().get_post_process_parameters()
        par = self.par

        if par.force:
            ppp.stretching_forces_array[(par.force, 0, 0)] = [self.first_dna_id]
            ppp.stretching_forces_array[(-par.force, 0, 0)] = [self.last_dna_id]
        else:
            ppp.end_points += [self.first_dna_id, self.last_dna_id]

        ppp.freeze_indices += [
            self.get_dna_id_from_list_index(
                get_closest(self.dna_full_list, self.smc.pos.r_lower_site[1]),
            ),  # closest to bottom -> r_lower_site[1]
            self.get_dna_id_from_list_index(
                get_closest(self.dna_full_list, self.smc.pos.r_middle_site[1]),
            ),  # closest to middle -> r_middle_site[1]
        ]

        ppp.dna_indices_list += self.dna_indices_list_get_all_dna()

        return ppp

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        return [self.last_dna_id]


class Folded(DnaConfiguration):
    def __init__(self, dna_groups, dna_parameters: DnaParameters, dna_center):
        super().__init__(dna_groups, dna_parameters)
        self.dna_center = dna_center

    @classmethod
    def get_dna_config(cls, dna_parameters: DnaParameters, r_lower_site, par) -> Folded:
        # place your DNA here, inside the SMC
        default_dna_pos = r_lower_site[1] + np.array([0, par.cutoff6, 0])

        # 1.
        [r_DNA], dna_center = dna_creator.get_dna_coordinates_twist(
            dna_parameters.nDNA, dna_parameters.DNA_bond_length, 17
        )

        # 2.
        # make sure SMC contains DNA
        goal = default_dna_pos
        start = np.array(
            [dna_center[0] + 10.0 * dna_parameters.DNA_bond_length, r_DNA[-1][1], 0]
        )
        shift = (goal - start).reshape(1, 3)
        r_DNA += shift

        return cls(dna_parameters.create_dna([r_DNA]), dna_parameters, dna_center)

    def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
        ppp = super().get_post_process_parameters()
        par = self.par

        if par.force:
            ppp.stretching_forces_array[(par.force, 0, 0)] = [
                self.first_dna_id,
                self.last_dna_id,
            ]
        else:
            ppp.end_points += [self.first_dna_id, self.last_dna_id]

        ppp.freeze_indices += [
            self.get_dna_id_from_list_index(
                get_closest(self.dna_full_list, self.smc.pos.r_lower_site[1]),
            ),  # closest to bottom -> r_lower_site[1]
            self.get_dna_id_from_list_index(
                get_closest(self.dna_full_list, self.smc.pos.r_middle_site[1]),
            ),  # closest to middle -> r_middle_site[1]
        ]

        ppp.dna_indices_list += self.dna_indices_list_get_dna_to(ratio=0.5)

        return ppp

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        # TODO: maybe add bead at halfway point (left most point)?
        return []


class RightAngle(DnaConfiguration):
    def __init__(self, dna_groups, dna_parameters: DnaParameters, dna_center):
        super().__init__(dna_groups, dna_parameters)
        self.dna_center = dna_center

    @classmethod
    def get_dna_config(
        cls, dna_parameters: DnaParameters, r_lower_site, par
    ) -> RightAngle:
        # place your DNA here, inside the SMC
        default_dna_pos = r_lower_site[1] + np.array([0, par.cutoff6, 0])

        # 1.
        [r_DNA], dna_center = dna_creator.get_dna_coordinates(
            dna_parameters.nDNA, dna_parameters.DNA_bond_length, 14, 10
        )

        # 2.
        # make sure SMC touches the DNA at the lower site (siteD)
        goal = default_dna_pos
        start = np.array(
            [dna_center[0] - 10.0 * dna_parameters.DNA_bond_length, dna_center[1], 0]
        )
        shift = (goal - start).reshape(1, 3)
        r_DNA += shift

        return cls(dna_parameters.create_dna([r_DNA]), dna_parameters, dna_center)

    def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
        ppp = super().get_post_process_parameters()
        par = self.par

        if par.force:
            ppp.stretching_forces_array[(0, par.force, 0)] = [self.first_dna_id]
            ppp.stretching_forces_array[(-par.force, 0, 0)] = [self.last_dna_id]
        else:
            ppp.end_points += [self.first_dna_id, self.last_dna_id]

        # find closest DNA bead to siteD
        # closest_DNA_index = get_closest(self.dna_groups[0].positions, r_lower_site[1])

        ppp.dna_indices_list += [(self.first_dna_id, self.get_percent_dna_id(0.5))]

        return ppp

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        return []


class Doubled(DnaConfiguration):
    def __init__(self, dna_groups, dna_parameters: DnaParameters, dna_center):
        super().__init__(dna_groups, dna_parameters)
        self.dna_center = dna_center

    @classmethod
    def get_dna_config(
        cls, dna_parameters: DnaParameters, r_lower_site, par
    ) -> Doubled:
        # place your DNA here, inside the SMC
        default_dna_pos = r_lower_site[1] + np.array([0, par.cutoff6, 0])

        # 1.
        r_DNA_list, dna_center = dna_creator.get_dna_coordinates_doubled(
            dna_parameters.nDNA, dna_parameters.DNA_bond_length, 24
        )

        # 2.
        # make sure SMC contains DNA
        goal = default_dna_pos
        start = np.array(
            [
                dna_center[0] + 30.0 * dna_parameters.DNA_bond_length,
                r_DNA_list[0][-1][1],
                0,
            ]
        )
        shift = (goal - start).reshape(1, 3)
        r_DNA_list[0] += shift
        r_DNA_list[1] += shift

        return cls(dna_parameters.create_dna(r_DNA_list), dna_parameters, dna_center)

    def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
        ppp = super().get_post_process_parameters()
        par = self.par

        # get dna beads to freeze
        for dna_grp in self.dna_groups:
            if par.force:
                ppp.stretching_forces_array[(par.force, 0, 0)] = [
                    (dna_grp, 0),
                    (dna_grp, -1),
                ]
            else:
                ppp.end_points += [(dna_grp, 0), (dna_grp, -1)]
            # TODO: fix for DOUBLED DNA, gives same bead twice
            ppp.freeze_indices += [
                (
                    dna_grp,
                    get_closest(dna_grp.positions, self.smc.pos.r_lower_site[1]),
                ),  # closest to bottom
                (
                    dna_grp,
                    get_closest(dna_grp.positions, self.smc.pos.r_middle_site[1]),
                ),  # closest to middle
            ]

        ppp.dna_indices_list += self.dna_indices_list_get_dna_to(ratio=0.5)

        return ppp

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        # TODO: see todo for folded config
        return []


@with_tether
class Obstacle(DnaConfiguration):
    def __init__(
        self,
        dna_groups,
        dna_parameters: DnaParameters,
        tether: Tether,
        dna_start_index: int,
    ):
        super().__init__(dna_groups, dna_parameters)
        self.tether = tether
        self.dna_start_index = dna_start_index

    @classmethod
    def get_dna_config(
        cls, dna_parameters: DnaParameters, r_lower_site, par
    ) -> Obstacle:
        # place your DNA here, inside the SMC
        default_dna_pos = r_lower_site[1] + np.array([0, par.cutoff6, 0])

        # 1.
        [r_DNA] = dna_creator.get_dna_coordinates_straight(
            dna_parameters.nDNA, dna_parameters.DNA_bond_length
        )

        # 2.
        # make sure SMC contains DNA
        goal = default_dna_pos
        dna_start_index = int(len(r_DNA) * 9 / 15)
        start = np.array(
            [
                r_DNA[dna_start_index][0] - 10.0 * dna_parameters.DNA_bond_length,
                r_DNA[dna_start_index][1],
                0,
            ]
        )
        shift = (goal - start).reshape(1, 3)
        r_DNA += shift

        dna_groups = dna_parameters.create_dna([r_DNA])

        dna_bead_to_tether_id = int(len(r_DNA) * 7.5 / 15)
        tether = Tether.create_tether(
            (dna_groups[0], dna_bead_to_tether_id),
            25,
            dna_parameters.DNA_bond_length,
            dna_parameters.DNA_mass,
            dna_parameters.bond,
            dna_parameters.ssangle,
            Tether.Obstacle(),
        )
        obstacle = Tether.get_obstacle(True, cls.inter_par, tether.group)
        tether.obstacle = obstacle
        # place the tether next to the DNA bead
        tether.move(r_DNA[dna_bead_to_tether_id] - tether.group.positions[-1])
        # move down a little
        tether.move(np.array([0, -dna_parameters.DNA_bond_length, 0], dtype=float))

        return cls(dna_groups, dna_parameters, tether, dna_start_index)

    def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
        ppp = super().get_post_process_parameters()
        par = self.par

        if par.force:
            ppp.stretching_forces_array[(par.force, 0, 0)] = [self.first_dna_id]
            ppp.stretching_forces_array[(-par.force, 0, 0)] = [self.last_dna_id]
        else:
            ppp.end_points += [self.first_dna_id, self.last_dna_id]

        ppp.dna_indices_list += [
            ((dna_grp, 0), (dna_grp, self.dna_start_index))
            for dna_grp in self.dna_groups
        ]

        return ppp

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        return [self.last_dna_id]


class Safety(DnaConfiguration):
    def __init__(
        self,
        dna_groups,
        dna_parameters: DnaParameters,
        dna_safety_belt_index,
    ):
        super().__init__(dna_groups, dna_parameters)
        self.dna_safety_belt_index = dna_safety_belt_index

    @classmethod
    def get_dna_config(cls, dna_parameters: DnaParameters, r_lower_site, par) -> Safety:
        # 1.
        [r_DNA], belt_location, dna_safety_belt_index, _ = (
            dna_creator.get_dna_coordinates_safety_belt(
                dna_parameters.nDNA, dna_parameters.DNA_bond_length
            )
        )

        # 2.
        # make sure SMC contains DNA
        shift = r_lower_site[1] - belt_location
        shift[1] -= 0.65 * par.cutoff6 + 0.5 * par.cutoff6  # TODO: if siteDup
        r_DNA += shift

        dna_groups = dna_parameters.create_dna([r_DNA])

        return cls(dna_groups, dna_parameters, dna_safety_belt_index)

    def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
        ppp = super().get_post_process_parameters()
        par = self.par

        if par.force:
            ppp.stretching_forces_array[(par.force, 0, 0)] = [self.first_dna_id]
            ppp.stretching_forces_array[(-par.force, 0, 0)] = [self.last_dna_id]
        else:
            ppp.end_points += [self.first_dna_id, self.last_dna_id]

        ppp.dna_indices_list += self.dna_indices_list_get_all_dna()

        # prevent breaking of safety belt
        # ppp.freeze_indices += [
        #     *[(self.dna_groups[0], self.dna_safety_belt_index + i) for i in range(-6, 6)],
        #     *[(self.smc.siteD_group, i) for i in range(len(self.smc.siteD_group.positions))]
        # ]

        return ppp

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        return [self.last_dna_id]


@with_tether
class ObstacleSafety(DnaConfiguration):
    def __init__(
        self,
        dna_groups,
        dna_parameters: DnaParameters,
        tether: Tether,
        dna_safety_belt_index: int,
    ):
        super().__init__(dna_groups, dna_parameters)
        self.tether = tether
        self.dna_safety_belt_index = dna_safety_belt_index

    @classmethod
    def get_dna_config(
        cls, dna_parameters: DnaParameters, r_lower_site, par
    ) -> ObstacleSafety:
        # 1.
        [r_DNA], belt_location, dna_safety_belt_index, dna_bead_to_tether_id = (
            dna_creator.get_dna_coordinates_safety_belt(
                dna_parameters.nDNA, dna_parameters.DNA_bond_length
            )
        )

        # 2.
        # make sure SMC contains DNA
        shift = r_lower_site[1] - belt_location
        shift[1] -= 0.65 * par.cutoff6 + 0.5 * par.cutoff6  # TODO: if siteDup
        r_DNA += shift

        dna_groups = dna_parameters.create_dna([r_DNA])

        tether = Tether.create_tether(
            (dna_groups[0], dna_bead_to_tether_id),
            35,
            dna_parameters.DNA_bond_length,
            dna_parameters.DNA_mass,
            dna_parameters.bond,
            dna_parameters.ssangle,
            Tether.Obstacle(),
        )
        obstacle = Tether.get_obstacle(True, cls.inter_par, tether.group)
        tether.obstacle = obstacle

        tether.move(r_DNA[dna_bead_to_tether_id] - tether.group.positions[-1])
        # move down a little
        tether.move(np.array([0, -dna_parameters.DNA_bond_length, 0], dtype=float))

        return cls(dna_groups, dna_parameters, tether, dna_safety_belt_index)

    def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
        ppp = super().get_post_process_parameters()
        par = self.par

        if par.force:
            ppp.stretching_forces_array[(par.force, 0, 0)] = [self.first_dna_id]
            ppp.stretching_forces_array[(-par.force, 0, 0)] = [self.last_dna_id]
        else:
            ppp.end_points += [self.first_dna_id, self.last_dna_id]

        ppp.dna_indices_list += self.dna_indices_list_get_all_dna()

        # prevent breaking of safety belt
        # ppp.freeze_indices += [
        #     *[(self.dna_groups[0], self.dna_safety_belt_index + i) for i in range(-6, 6)],
        #     *[(self.smc.siteD_group, i) for i in range(len(self.smc.siteD_group.positions))]
        # ]

        return ppp

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        return [self.last_dna_id]


@with_tether
class AdvancedObstacleSafety(DnaConfiguration):
    def __init__(
        self,
        dna_groups,
        dna_parameters: DnaParameters,
        tether: Tether,
        dna_safety_belt_index: int,
    ):
        super().__init__(dna_groups, dna_parameters)
        self.tether = tether
        self.dna_safety_belt_index = dna_safety_belt_index

    @classmethod
    def get_dna_config(
        cls, dna_parameters: DnaParameters, r_lower_site, par
    ) -> AdvancedObstacleSafety:
        # 1.
        # [rDNA], belt_location, dna_safety_belt_index, dna_bead_to_tether_id = dna_creator.get_dna_coordinates_advanced_safety_belt(dna_parameters.nDNA, dna_parameters.DNAbondLength)
        [r_DNA], belt_location, dna_safety_belt_index, dna_bead_to_tether_id = (
            dna_creator.get_dna_coordinates_advanced_safety_belt_plus_loop(
                dna_parameters.nDNA, dna_parameters.DNA_bond_length
            )
        )

        # 2.
        # make sure SMC contains DNA
        shift = r_lower_site[1] - belt_location
        shift[1] -= 1.35 * par.cutoff6 + 0.5 * par.cutoff6  # TODO: if siteDup
        r_DNA += shift

        dna_groups = dna_parameters.create_dna([r_DNA])

        tether = Tether.create_tether(
            (dna_groups[0], dna_bead_to_tether_id),
            35,
            dna_parameters.DNA_bond_length,
            dna_parameters.DNA_mass,
            dna_parameters.bond,
            dna_parameters.ssangle,
            Tether.Obstacle(),
        )
        obstacle = Tether.get_obstacle(True, cls.inter_par, tether.group)
        tether.obstacle = obstacle

        # place the tether next to the DNA bead
        tether.move(r_DNA[dna_bead_to_tether_id] - tether.group.positions[-1])
        # move down a little
        tether.move(np.array([0, -dna_parameters.DNA_bond_length, 0], dtype=float))

        return cls(dna_groups, dna_parameters, tether, dna_safety_belt_index)

    def get_post_process_parameters(self) -> DnaConfiguration.PostProcessParameters:
        ppp = super().get_post_process_parameters()
        par = self.par

        if par.force:
            ppp.stretching_forces_array[(par.force, 0, 0)] = [self.first_dna_id]
            ppp.stretching_forces_array[(-par.force, 0, 0)] = [self.last_dna_id]
        else:
            ppp.end_points += [self.first_dna_id, self.last_dna_id]

        ppp.dna_indices_list += self.dna_indices_list_get_all_dna()

        # prevent breaking of safety belt
        # ppp.freeze_indices += [
        #     *[(self.dna_groups[0], self.dna_safety_belt_index + i) for i in range(-6, 6)],
        #     *[(self.smc.siteD_group, i) for i in range(len(self.smc.siteD_group.positions))]
        # ]

        return ppp

    def get_stopper_ids(self) -> List[AtomIdentifier]:
        return [self.last_dna_id]



from pathlib import Path
from typing import Any, List

import numpy as np
import numpy.typing as npt

from smc_lammps.generate.generator import AtomIdentifier, Generator


def get_project_root() -> Path:
    return Path(__file__).parent.parent


def create_phase(
    generator: Generator, phase_path: Path, options: List[Generator.DynamicCoeffs]
):
    """creates a file containing coefficients to dynamically load in LAMMPS scripts"""

    def apply(function, file, list_of_args: List[Any]):
        for args in list_of_args:
            function(file, args)

    with open(phase_path, "w", encoding="utf-8") as phase_file:
        apply(generator.write_script_bai_coeffs, phase_file, options)


def get_closest(array, position) -> int:
    """returns the index of the array that is closest to the given position"""

    distances = np.linalg.norm(array - position, axis=1)
    return int(np.argmin(distances))


def pos_from_id(atom_id: AtomIdentifier) -> npt.NDArray[np.float32]:
    """get the position of an atom from its identifier"""
    return np.copy(atom_id[0].positions[atom_id[1]])



#### DO NOT EDIT THIS BLOCK ####
from smc_lammps.generate.default_parameters import Parameters

p = Parameters()
#### END OF BLOCK ####

## Define your parameters using p.key = value
## See src/smc_lammps/generate/default_parameters.py for all parameters

## Below are some example parameters


# Radius of lower circular-arc compartment (nm)
p.kleisin_radius = 4.5

# Amount of DNA
p.N = 300

# Set number of cycles directly
# p.cycles = 20

# Or set max_steps
p.cycles = None
total_steps_per_cycle = p.steps_APO + p.steps_ADP + p.steps_ATP
p.max_steps = 20 * total_steps_per_cycle

# Hinge acts as one rigid part
p.rigid_hinge = True

# Small obstacle on DNA at tether location
p.add_RNA_polymerase = False
# p.RNA_polymerase_type = 1
# p.RNA_polymerase_size = 5.0

# Apply a weak force on the ends of the DNA
p.force = 0.05

# Choose a configuration
# p.dna_config = "advanced_obstacle_safety"
# p.dna_config = "obstacle_safety"
p.dna_config = "obstacle"
# p.dna_config = "line"

# Prevent SMC from falling off the DNA
p.add_stopper_bead = True

# Enable or disable charges
p.use_charges = False



# Copyright (c) 2021 Stefanos Nomidis
# Copyright (c) 2022 Arwin Goossens
# Copyright (c) 2024 Lucas Dooms

import math
from collections.abc import Sequence
from pathlib import Path
from runpy import run_path
from sys import argv, maxsize
from typing import Any, List

import numpy as np
from numpy.random import default_rng

from smc_lammps.generate.default_parameters import Parameters
from smc_lammps.generate.generator import (
    AtomIdentifier,
    AtomType,
    BAI_Kind,
    BAI_Type,
    Generator,
    MoleculeId,
    PairWise,
)
from smc_lammps.generate.structures.dna import dna
from smc_lammps.generate.structures.smc.smc import SMC
from smc_lammps.generate.structures.smc.smc_creator import SMC_Creator
from smc_lammps.generate.util import create_phase, get_closest

if len(argv) < 2:
    raise ValueError("Provide a folder path")
path = Path(argv[1])

par = run_path((path / "parameters.py").as_posix())["p"]

assert isinstance(par, Parameters)

nDNA = par.N
bases_per_bead = par.n


#################################################################################
#                               Other parameters                                #
#################################################################################


# Simulation temperature (K)
T = par.T

# Boltzmann's constant (pN nm / K)
kB = par.kB

kBT = kB * T


#################################### Masses #####################################


#######
# DNA #
#######

# Mass per base pair (ag)
basepair_mass = 2 * 3.1575 * 5.24e-4

# Effective bead mass (ag)
DNA_bead_mass = bases_per_bead * basepair_mass


#######
# SMC #
#######

# Total mass of SMC protein (ag)
SMC_total_mass = 0.25


#################################### Lengths ####################################


################
# Interactions #
################

# DNA-DNA repulsion radius (nm)
radius_DNA_DNA = 3.5


#######
# DNA #
#######

# Bending stiffness (nm)
DNA_persistence_length = 50.0
ssDNA_persistence_length = 1.0

# Base pair step (nm)
basepair_size = 0.34

# Effective bond length = DNA bead size (nm)
DNA_bond_length = bases_per_bead * basepair_size

# Total length of DNA (nm)
DNA_total_length = DNA_bond_length * nDNA


#######
# SMC #
#######

# Desirable SMC spacing (radius of 1 SMC bead is R = intRadSMCvsDNA)
# Equal to R:   Minimum diameter = sqrt(3)    = 1.73 R
# Equal to R/2: Minimum diameter = sqrt(15)/2 = 1.94 R
SMC_spacing = par.sigma_SMC_DNA / 2


##################
# Simulation box #
##################


# Width of cubic simulation box (nm)
box_width = 2 * DNA_total_length


################################## Interactions #################################


###########
# DNA-DNA #
###########

sigma_DNA_DNA = radius_DNA_DNA
epsilon_DNA_DNA = par.epsilon3
rcut_DNA_DNA = sigma_DNA_DNA * 2 ** (1 / 6)


###########
# SMC-DNA #
###########

sigma_SMC_DNA = par.sigma_SMC_DNA
epsilon_SMC_DNA = par.epsilon3
rcut_SMC_DNA = sigma_SMC_DNA * 2 ** (1 / 6)


#############
# Sites-DNA #
#############

# Sigma of LJ attraction (same as those of the repulsive SMC sites)
sigma_upper_site_DNA = sigma_SMC_DNA

# Cutoff distance of LJ attraction
rcut_upper_site_DNA = par.cutoff6

# Epsilon parameter of LJ attraction
epsilon_upper_site_DNA = par.epsilon6

interaction_parameters = dna.InteractionParameters(
    sigma_DNA_DNA=sigma_DNA_DNA,
    epsilon_DNA_DNA=epsilon_DNA_DNA,
    rcut_DNA_DNA=rcut_DNA_DNA,
    sigma_SMC_DNA=sigma_SMC_DNA,
    epsilon_SMC_DNA=epsilon_SMC_DNA,
    rcut_SMC_DNA=rcut_SMC_DNA,
    sigma_upper_site_DNA=sigma_upper_site_DNA,
    rcut_lower_site_DNA=rcut_upper_site_DNA,
    epsilon_upper_site_DNA=epsilon_upper_site_DNA,
)

# Even More Parameters


# Relative bond fluctuations
bond_fluctuation_DNA = 1e-2
bond_fluctuation_SMC = 1e-2
# bond_fluctuation_hinge = 0.5 # large fluctuations to allow tether passing
bond_fluctuation_hinge = 3e-2  # small fluctuations

# Maximum relative bond extension (units of rest length)
bond_max_extension = 1.0

# Spring constant obeying equilibrium relative bond fluctuations
k_bond_DNA = 3 * kBT / (DNA_bond_length * bond_fluctuation_DNA) ** 2
k_bond_SMC = 3 * kBT / (SMC_spacing * bond_fluctuation_SMC) ** 2
k_bond_hinge = 3 * kBT / (SMC_spacing * bond_fluctuation_hinge) ** 2


# Maximum bond length
max_bond_length_DNA = DNA_bond_length * bond_max_extension
max_bond_length_SMC = SMC_spacing * bond_max_extension

# DNA bending rigidity
k_angle_DNA = DNA_persistence_length * kBT / DNA_bond_length
k_angle_ssDNA = ssDNA_persistence_length * kBT / DNA_bond_length


#################################################################################
#                                 Start Setup                                   #
#################################################################################


dna.DnaConfiguration.set_parameters(par, interaction_parameters)
dna_config_class = dna.DnaConfiguration.str_to_config(par.dna_config)


#################################################################################
#                                 SMC complex                                   #
#################################################################################


smc_creator = SMC_Creator(
    SMC_spacing=SMC_spacing,
    #
    upper_site_v=4.0,
    upper_site_h=2.0,
    middle_site_v=1.0,
    middle_site_h=2.0,
    lower_site_v=0.5,
    lower_site_h=2.0,
    #
    arm_length=par.arm_length,
    bridge_width=par.bridge_width,
    hinge_radius=par.hinge_radius,
    # SMCspacing half of the minimal required spacing of ssDNA
    # so between 2*SMCspacing and 4*SMCspacing should
    # allow ssDNA passage but not dsDNA
    hinge_opening=2.2 * SMC_spacing,
    #
    kleisin_radius=par.kleisin_radius,
    folding_angle_APO=par.folding_angle_APO,
)

rot_vec = (
    np.array([0.0, 0.0, -np.deg2rad(42)])
    if dna_config_class is dna.AdvancedObstacleSafety
    else None
)
smc_positions = smc_creator.get_smc(
    lower_site_points_down=False,
    # dnaConfigClass in {dna.ObstacleSafety, dna.AdvancedObstacleSafety},
    extra_rotation=rot_vec,
)


#################################################################################
#                                     DNA                                       #
#################################################################################

# set DNA bonds, angles, and mass
mol_DNA = MoleculeId.get_next()
dna_bond = BAI_Type(
    BAI_Kind.BOND,
    "fene/expand",
    f"{k_bond_DNA} {max_bond_length_DNA} {0.0} {0.0} {DNA_bond_length}\n",
)
dna_angle = BAI_Type(BAI_Kind.ANGLE, "cosine", f"{k_angle_DNA}\n")
ssdna_angle = BAI_Type(BAI_Kind.ANGLE, "cosine", f"{k_angle_ssDNA}\n")
dna_type = AtomType(DNA_bead_mass)

dna_parameters = dna.DnaParameters(
    nDNA=nDNA,
    DNA_bond_length=DNA_bond_length,
    DNA_mass=DNA_bead_mass,
    type=dna_type,
    mol_DNA=mol_DNA,
    bond=dna_bond,
    angle=dna_angle,
    ssangle=ssdna_angle,
)
dna_config = dna_config_class.get_dna_config(
    dna_parameters, smc_positions.r_lower_site, par
)

#################################################################################
#                                Print to file                                  #
#################################################################################

# Divide total mass evenly among the segments
mSMC = smc_creator.get_mass_per_atom(SMC_total_mass)


# SET UP DATAFILE GENERATOR
gen = Generator()
gen.set_system_size(box_width)
gen.use_charges = par.use_charges
if gen.use_charges:
    # prevents inf/nan in coul calculations
    shift_rng = default_rng(par.seed)
    gen.random_shift = lambda: shift_rng.normal(0, 1e-6 * DNA_bond_length, (3,))

smc_1 = SMC(
    use_rigid_hinge=par.rigid_hinge,
    pos=smc_positions,
    #
    t_arms_heads_kleisin=AtomType(mSMC),
    t_hinge=AtomType(mSMC),
    t_atp=AtomType(mSMC),
    t_upper_site=AtomType(mSMC),
    t_middle_site=AtomType(mSMC),
    t_lower_site=AtomType(mSMC),
    t_ref_site=AtomType(mSMC),
    #
    k_bond=k_bond_SMC,
    k_hinge=k_bond_hinge,
    max_bond_length=max_bond_length_SMC,
    #
    k_elbow=par.elbows_stiffness * kBT,
    k_arm=par.arms_stiffness * kBT,
    #
    k_align_site=par.site_stiffness * kBT,
    k_fold=par.folding_stiffness * kBT,
    k_asymmetry=par.asymmetry_stiffness * kBT,
    #
    bridge_width=par.bridge_width,
    arm_length=par.arm_length,
    hinge_radius=par.hinge_radius,
    arms_angle_ATP=par.arms_angle_ATP,
    folding_angle_ATP=par.folding_angle_ATP,
    folding_angle_APO=par.folding_angle_APO,
)

dna_config.set_smc(smc_1)

extra_mols_smc: List[int] = []
extra_mols_dna: List[int] = []

if par.add_RNA_polymerase:
    mol_bead = MoleculeId.get_next()
    bead_type = AtomType(10.0 * DNA_bead_mass)
    bead_size = par.RNA_polymerase_size

    if par.RNA_polymerase_type == 0:
        bead_bond = BAI_Type(BAI_Kind.BOND, "harmonic", f"{k_bond_DNA} {bead_size}\n")
        bead_angle = dna_angle
        extra_mols_dna.append(mol_bead)
    elif par.RNA_polymerase_type == 1:
        bead_bond = None
        bead_angle = None
        extra_mols_smc.append(mol_bead)
    else:
        raise ValueError(f"unknown RNA_polymerase_type, {par.RNA_polymerase_type}")

    if hasattr(dna_config, "tether"):
        dna_id = dna_config.tether.dna_tether_id
    else:
        dna_id = (
            dna_config.dna_groups[0],
            int(len(dna_config.dna_groups[0].positions) // 2),
        )
    dna_config.add_bead_to_dna(
        bead_type, mol_bead, dna_id, bead_bond, bead_angle, bead_size
    )

    if bead_bond is None:
        gen.molecule_override[dna_id] = mol_bead

if par.spaced_beads_interval is not None:
    spaced_bead_type = AtomType(DNA_bead_mass)

    # get spacing
    start_id = par.spaced_beads_interval
    stop_id = get_closest(
        dna_config.dna_groups[0].positions, smc_positions.r_lower_site[1]
    )
    spaced_bead_ids = list(range(start_id, stop_id, par.spaced_beads_interval))
    spaced_bead_ids += list(
        range(
            stop_id + par.spaced_beads_interval,
            len(dna_config.dna_groups[0].positions),
            par.spaced_beads_interval,
        )
    )

    for dna_id in spaced_bead_ids:
        mol_spaced_bead = MoleculeId.get_next()
        extra_mols_smc.append(mol_spaced_bead)
        dna_id = (dna_config.dna_groups[0], dna_id)
        dna_config.add_bead_to_dna(
            spaced_bead_type,
            mol_spaced_bead,
            dna_id,
            None,
            None,
            par.spaced_beads_size,
        )
        gen.molecule_override[dna_id] = mol_spaced_bead

if par.add_stopper_bead:
    mol_stopper = MoleculeId.get_next()
    extra_mols_smc.append(mol_stopper)
    stopper_type = AtomType(0.01 * DNA_bead_mass)
    stopper_size = 25.0

    stopper_ids = dna_config.get_stopper_ids()
    for dna_id in stopper_ids:
        dna_config.add_bead_to_dna(
            stopper_type, mol_stopper, dna_id, None, None, stopper_size
        )
        gen.molecule_override[dna_id] = mol_stopper


gen.atom_groups += [
    *dna_config.get_all_groups(),
    *smc_1.get_groups(),
]


# Pair coefficients
pair_inter = PairWise(
    "PairIJ Coeffs # hybrid\n\n", "lj/cut {} {} {}\n", [0.0, 0.0, 0.0]
)

dna_config.add_interactions(pair_inter)
smc_1.add_repel_interactions(
    pair_inter, epsilon_SMC_DNA * kBT, sigma_SMC_DNA, rcut_SMC_DNA
)

# soft interactions
pair_soft_inter = PairWise("PairIJ Coeffs # hybrid\n\n", "soft {} {}\n", [0.0, 0.0])

gen.pair_interactions.append(pair_inter)
gen.pair_interactions.append(pair_soft_inter)
if gen.use_charges:
    gen.pair_interactions.append(
        PairWise("PairIJ Coeffs # hybrid\n\n", "coul/debye {}\n", [""])
    )

# Interactions that change for different phases of SMC
bridge_off = Generator.DynamicCoeffs(None, "lj/cut 0 0 0\n", [dna_type, smc_1.t_atp])
bridge_on = Generator.DynamicCoeffs(
    None,
    f"lj/cut {epsilon_SMC_DNA * kBT} {par.sigma} {par.sigma * 2 ** (1 / 6)}\n",
    [dna_type, smc_1.t_atp],
)

bridge_soft_off = Generator.DynamicCoeffs(None, "soft 0 0\n", [dna_type, smc_1.t_atp])
bridge_soft_on = Generator.DynamicCoeffs(
    None,
    f"soft {epsilon_SMC_DNA * kBT} {par.sigma * 2 ** (1 / 6)}\n",
    [dna_type, smc_1.t_atp],
)

hinge_attraction_off = Generator.DynamicCoeffs(
    None, "lj/cut 0 0 0\n", [dna_type, smc_1.t_upper_site]
)
hinge_attraction_on = Generator.DynamicCoeffs(
    None,
    f"lj/cut {par.epsilon4 * kBT} {par.sigma} {par.cutoff4}\n",
    [dna_type, smc_1.t_upper_site],
)

if False:  # isinstance(dnaConfig, (dna.ObstacleSafety, dna.AdvancedObstacleSafety))
    # always keep site on
    lower_site_off = Generator.DynamicCoeffs(
        None,
        f"lj/cut {par.epsilon6 * kBT} {par.sigma} {par.cutoff6}\n",
        [dna_type, smc_1.t_lower_site],
    )
else:
    lower_site_off = Generator.DynamicCoeffs(
        None, "lj/cut 0 0 0\n", [dna_type, smc_1.t_lower_site]
    )
lower_site_on = Generator.DynamicCoeffs(
    None,
    f"lj/cut {par.epsilon6 * kBT} {par.sigma} {par.cutoff6}\n",
    [dna_type, smc_1.t_lower_site],
)

middle_site_off = Generator.DynamicCoeffs(
    None, "lj/cut 0 0 0\n", [dna_type, smc_1.t_middle_site]
)
middle_site_on = Generator.DynamicCoeffs(
    None,
    f"lj/cut {par.epsilon5 * kBT} {par.sigma} {par.cutoff5}\n",
    [dna_type, smc_1.t_middle_site],
)

middle_site_soft_off = Generator.DynamicCoeffs(
    None, "soft 0 0\n", [dna_type, smc_1.t_middle_site]
)
middle_site_soft_on = Generator.DynamicCoeffs(
    None,
    "soft " + f"{par.epsilon5 * kBT} {par.sigma * 2 ** (1 / 6)}\n",
    [dna_type, smc_1.t_middle_site],
)

gen.bais += [*smc_1.get_bonds(smc_creator.hinge_opening), *dna_config.get_bonds()]

gen.bais += smc_1.get_angles()

gen.bais += smc_1.get_impropers()

# Override molecule ids to form rigid safety-belt bond
if isinstance(
    dna_config, (dna.ObstacleSafety, dna.AdvancedObstacleSafety, dna.Safety)
):  # TODO
    safety_index = dna_config.dna_safety_belt_index
    gen.molecule_override[(dna_config.dna_groups[0], safety_index)] = (
        smc_1.mol_lower_site
    )
    # add neighbors to prevent rotation
    # gen.molecule_override[(dnaConfig.dna_groups[0], safety_index - 1)] = smc_1.mol_lower_site
    # gen.molecule_override[(dnaConfig.dna_groups[0], safety_index + 1)] = smc_1.mol_lower_site

with open(path / "datafile_coeffs", "w", encoding="utf-8") as datafile:
    gen.write_coeffs(datafile)

with open(path / "datafile_positions", "w", encoding="utf-8") as datafile:
    gen.write_positions_and_bonds(datafile)

with open(path / "styles", "w", encoding="utf-8") as stylesfile:
    stylesfile.write(gen.get_atom_style_command())
    stylesfile.write(gen.get_BAI_styles_command())
    pair_style = "pair_style hybrid/overlay lj/cut $(3.5) soft $(3.5)"
    if gen.use_charges:
        pair_style += " coul/debye $(1.0/5.0) $(7.5)"
    stylesfile.write(pair_style)

# VMD visualization of initial configuration
with open(path / "vmd.tcl", "w", encoding="utf-8") as vmdfile:
    vmdfile.write(f"topo readlammpsdata {path / 'datafile_positions'}")

#################################################################################
#                                Phases of SMC                                  #
#################################################################################

# make sure the directory exists
states_path = path / "states"
states_path.mkdir(exist_ok=True)

create_phase(
    gen,
    states_path / "adp_bound",
    [
        bridge_off,
        hinge_attraction_on,
        middle_site_off,
        lower_site_off,
        smc_1.arms_open,
        smc_1.kleisin_unfolds1,
        smc_1.kleisin_unfolds2,
    ],
)

create_phase(
    gen,
    states_path / "apo",
    [
        bridge_off,
        hinge_attraction_off,
        middle_site_off,
        lower_site_on,
        smc_1.arms_close,
        smc_1.kleisin_unfolds1,
        smc_1.kleisin_unfolds2,
    ],
)
# gen.write_script_bai_coeffs(adp_bound_file, BAI_Kind.ANGLE, "{} harmonic " + f"{angle3kappa} {angle3angleAPO2}\n", angle_t3)   # Arms close MORE

create_phase(
    gen,
    states_path / "atp_bound_1",
    [
        bridge_soft_on,
        middle_site_soft_on,
    ],
)

create_phase(
    gen,
    states_path / "atp_bound_2",
    [
        bridge_soft_off,
        middle_site_soft_off,
        bridge_on,
        hinge_attraction_on,
        middle_site_on,
        lower_site_on,
        smc_1.arms_open,
        smc_1.kleisin_folds1,
        smc_1.kleisin_folds2,
    ],
)


#################################################################################
#                           Print to post processing                            #
#################################################################################

ppp = dna_config.get_post_process_parameters()

with open(path / "post_processing_parameters.py", "w", encoding="utf-8") as file:
    file.write(
        "# use to form plane of SMC arms\n"
        f"top_left_bead_id = {gen.get_atom_index((smc_1.arm_ul_grp, -1))}\n"
        f"top_right_bead_id = {gen.get_atom_index((smc_1.arm_ur_grp, 0))}\n"
        f"left_bead_id = {gen.get_atom_index((smc_1.arm_dl_grp, -1))}\n"
        f"right_bead_id = {gen.get_atom_index((smc_1.arm_dr_grp, 0))}\n"
        f"middle_left_bead_id = {gen.get_atom_index((smc_1.atp_grp, 0))}\n"
        f"middle_right_bead_id = {gen.get_atom_index((smc_1.atp_grp, -1))}\n"
    )
    file.write("\n")
    dna_indices_list = [
        (gen.get_atom_index(atomId1), gen.get_atom_index(atomId2))
        for (atomId1, atomId2) in ppp.dna_indices_list
    ]
    file.write(
        "# list of (min, max) of DNA indices for separate pieces to analyze\n"
        f"dna_indices_list = {dna_indices_list}\n"
    )
    file.write("\n")
    kleisin_ids_list = [
        gen.get_atom_index((smc_1.hk_grp, i))
        for i in range(len(smc_1.hk_grp.positions))
    ]
    file.write(
        f"# use to form plane of SMC kleisin\nkleisin_ids = {kleisin_ids_list}\n"
    )
    file.write("\n")
    file.write(f"dna_spacing = {max_bond_length_DNA}\n")
    file.write("\n")
    file.write(
        f"DNA_types = {list(set(grp.type.index for grp in dna_config.dna_groups))}\n"
    )
    file.write(
        f"SMC_types = {list(set(grp.type.index for grp in smc_1.get_groups()))}\n"
    )


#################################################################################
#                           Print to parameterfile                              #
#################################################################################


def atomIds_to_LAMMPS_ids(lst: Sequence[AtomIdentifier]) -> List[int]:
    return [gen.get_atom_index(atomId) for atomId in lst]


def get_variables_for_lammps() -> List[str]:
    """returns variable names that are needed in LAMMPS script"""
    return [
        "T",
        "gamma",
        "seed",
        "output_steps",
        "epsilon3",
        "sigma",
        "timestep",
        "smc_force",
    ]


def list_to_space_str(lst: Sequence[Any], surround="") -> str:
    """turn list into space separated string
    example: [1, 2, 6] -> 1 2 6"""
    return " ".join([surround + str(val) + surround for val in lst])


def prepend_or_empty(string: str, prepend: str) -> str:
    """prepend something if the string is non-empty
    otherwise replace it with the string "empty"."""
    if string:
        return prepend + string
    return "empty"


def get_string_def(name: str, value: str) -> str:
    """define a LAMMPS string"""
    return f'variable {name} string "{value}"\n'


def get_universe_def(name: str, values: Sequence[Any]) -> str:
    """define a LAMMPS universe"""
    return f"""variable {name} universe {list_to_space_str(values, surround='"')}\n"""


def get_index_def(name: str, values: Sequence[Any]) -> str:
    """define a LAMMPS universe"""
    return f"variable {name} index {list_to_space_str(values)}\n"


def get_times(
    apo: int, atp1: int, atp2: int, adp: int, rng_gen: np.random.Generator
) -> List[int]:
    # get run times for each SMC state
    # APO -> ATP1 -> ATP2 -> ADP -> ...

    def mult(x):
        # use 1.0 to get (0, 1] lower exclusive
        return -x * np.log(1.0 - rng_gen.uniform())

    return [math.ceil(mult(x)) for x in (apo, atp1, atp2, adp)]


def get_times_with_max_steps(
    parameters: Parameters, rng_gen: np.random.Generator
) -> List[int]:
    run_steps = []

    def none_to_max(x):
        if x is None:
            return maxsize  # very large number!
        return x

    cycles_left = none_to_max(parameters.cycles)
    max_steps = none_to_max(parameters.max_steps)

    cum_steps = 0
    while True:  # use do while loop since run_steps should not be empty
        new_times = get_times(
            parameters.steps_APO,
            10000,
            parameters.steps_ATP,
            parameters.steps_ADP,
            rng_gen,
        )
        run_steps += new_times

        cum_steps += sum(new_times)
        cycles_left -= 1

        if cycles_left <= 0 or cum_steps >= max_steps:
            break

    return run_steps


with open(path / "parameterfile", "w", encoding="utf-8") as parameterfile:
    parameterfile.write("# LAMMPS parameter file\n\n")

    # change seed if arg 2 provided
    if len(argv) > 2:
        seed_overwrite = int(argv[2])
        par.seed = seed_overwrite
    params = get_variables_for_lammps()
    for key in params:
        parameterfile.write(f"variable {key} equal {getattr(par, key)}\n\n")

    # write molecule ids
    # NOTE: indices are allowed to be the same, LAMMPS will ignore duplicates
    parameterfile.write(
        get_string_def(
            "DNA_mols",
            list_to_space_str(
                list(
                    set(grp.molecule_index for grp in dna_config.get_all_groups())
                    - set(
                        grp.molecule_index for grp in dna_config.beads
                    )  # do not include RNA beads
                )
                + extra_mols_dna
            ),
        )
    )
    parameterfile.write(
        get_string_def(
            "SMC_mols",
            list_to_space_str(smc_1.get_molecule_ids() + extra_mols_smc),
        )
    )

    parameterfile.write("\n")

    # turn into LAMMPS indices
    end_points_LAMMPS = atomIds_to_LAMMPS_ids(ppp.end_points)
    parameterfile.write(
        get_string_def(
            "dna_end_points",
            prepend_or_empty(list_to_space_str(end_points_LAMMPS), "id "),
        )
    )

    # turn into LAMMPS indices
    freeze_indices_LAMMPS = atomIds_to_LAMMPS_ids(ppp.freeze_indices)
    parameterfile.write(
        get_string_def(
            "indices", prepend_or_empty(list_to_space_str(freeze_indices_LAMMPS), "id ")
        )
    )

    if isinstance(
        dna_config, (dna.Obstacle, dna.ObstacleSafety, dna.AdvancedObstacleSafety)
    ) and isinstance(dna_config.tether.obstacle, dna.Tether.Wall):
        parameterfile.write(
            f"variable wall_y equal {dna_config.tether.group.positions[0][1]}\n"
        )

        excluded = [
            gen.get_atom_index((dna_config.tether.group, 0)),
            gen.get_atom_index((dna_config.tether.group, 1)),
        ]
        parameterfile.write(
            get_string_def(
                "excluded", prepend_or_empty(list_to_space_str(excluded), "id ")
            )
        )

    # forces
    stretching_forces_array_LAMMPS = {
        key: atomIds_to_LAMMPS_ids(val)
        for key, val in ppp.stretching_forces_array.items()
    }
    if stretching_forces_array_LAMMPS:
        parameterfile.write(
            f"variable stretching_forces_len equal {len(stretching_forces_array_LAMMPS)}\n"
        )
        sf_ids = [
            prepend_or_empty(list_to_space_str(lst), "id ")
            for lst in stretching_forces_array_LAMMPS.values()
        ]
        parameterfile.write(get_universe_def("stretching_forces_groups", sf_ids))
        sf_forces = [
            list_to_space_str(tup) for tup in stretching_forces_array_LAMMPS.keys()
        ]
        parameterfile.write(get_universe_def("stretching_forces", sf_forces))

    # obstacle, if particle
    if hasattr(dna_config, "tether") and isinstance(
        dna_config.tether.obstacle, dna.Tether.Gold
    ):
        obstacle_lammps_id = gen.get_atom_index((dna_config.tether.obstacle.group, 0))
        parameterfile.write(f"variable obstacle_id equal {obstacle_lammps_id}\n")

    parameterfile.write("\n")

    # get run times for each SMC state
    # APO -> ATP1 -> ATP2 -> ADP -> ...
    rng = default_rng(par.seed)
    runtimes = get_times_with_max_steps(par, rng)

    parameterfile.write(get_index_def("runtimes", runtimes))



from rich.console import Console
from rich.panel import Panel
from rich.text import Text

console = Console()


def warn(message):
    """Prints the warning message in a panel."""
    console.print(
        Panel(
            Text(message, style="bold yellow"),
            title="⚠️ WARNING",
            title_align="left",
            border_style="yellow",
            expand=False,
        )
    )



# Copyright (c) 2024 Lucas Dooms

from glob import glob
from pathlib import Path
from sys import argv
from typing import List

import numpy as np


def get_npz_files_from_args(args: List[str]):
    files = []

    # replace glob patterns
    matches = []
    for arg in args:
        matches += glob(arg)

    for match in matches:
        match = Path(match)
        if match.is_dir():
            for npzfile in match.glob("*.npz"):
                files.append(str(npzfile))
        else:
            if match.suffix != ".npz":
                print(f"WARNING: entered non npz file: {match}")
            files.append(str(match))

    return files


def process(files: List[str]):
    files = get_npz_files_from_args(files)
    if not files:
        raise Exception("did not receive files to process")

    steps_array = []
    indices_array = []
    for file in files:
        data = np.load(file)
        steps_array.append(data["steps"])
        indices_array.append(data["ids"])

    shortest_steps_length = min([len(x) for x in steps_array])
    steps_array = [steps[:shortest_steps_length] for steps in steps_array]
    steps = steps_array[0]
    assert(all([np.array_equal(steps, others) for others in steps_array]))

    for i in range(len(indices_array)):
        indices_array[i] = indices_array[i][:shortest_steps_length]

    indices_array = np.array(indices_array).transpose()

    def custom_average(arr):
        return np.average(arr) if arr.size else -1

    averages = np.array([
        custom_average(indices[indices != -1]) # ignore -1, these are due to issues in process_displacement.py
        for indices in indices_array
    ])

    import matplotlib.pyplot as plt
    
    plt.figure(figsize=(8, 6), dpi=144)
    plt.title("Averaged index of DNA bead inside SMC loop in time")
    plt.xlabel("time")
    plt.ylabel(f"Average DNA bead index ({len(files)} samples)")
    plt.scatter(steps, averages, s=0.5)
    plt.savefig("average_bead_id_in_time.png")

if __name__ == "__main__":
    argv = argv[1:]
    if not argv:
        raise Exception("Please provide glob patterns of npz files or folders containing them")
    process(argv)
